"""This module contains utilities to run stages in "Match-revise" mode.

Match-revise mode allows one to dynamically select a set of filters and different configuration parameters
such that a given stage will run with the corresponding parameters on each of the sets of data in the input
dataframe that matches the given filter.
"""

import logging
from enum import StrEnum, auto
from functools import wraps

import hydra
import polars as pl
from meds import DataSchema
from omegaconf import DictConfig, ListConfig

from ..dataframe import DF_T
from ..parser import is_matcher, matcher_to_expr
from .compute_fn import ANY_COMPUTE_FN_T, COMPUTE_FN_T, bind_compute_fn

logger = logging.getLogger(__name__)

MATCH_REVISE_KEY = "_match_revise"
MATCHER_KEY = "_matcher"
MATCH_REVISE_MODE_KEY = "_match_revise_mode"


class MatchReviseMode(StrEnum):
    """The different modes for match and revise operations.

    Future modes to be considered, match and add, multi-match and add, filter and revise, multi-filter and
    revise.

    Attributes:
        MATCH_AND_REVISE: The match and revise mode, which iterates through the list of matcher/function pairs
            and filters the input DataFrame for rows that match the matcher and applies a local compute
            function to the filtered DataFrame. The DataFrame to be matched in future iterations is restricted
            to only those rows that have not yet been matched. The unmatched dataframe at the end of the
            operation is concatenated with the outputs of all intermediate dataframes.
        MULTI_MATCH_AND_REVISE: The match and revise mode, which iterates through the list of matcher/function
            pairs and filters the input DataFrame for rows that match the matcher and applies a local compute
            function to the filtered DataFrame. The DataFrame to be matched in future iterations is the entire
            raw dataframe, including rows that have already been matched. The portion of the dataframe that
            didn't match anything on input is concatenated with the outputs of all intermediate dataframes.
    """

    MATCH_AND_REVISE = auto()
    MULTI_MATCH_AND_REVISE = auto()


def is_match_revise(stage_cfg: DictConfig) -> bool:
    """Check if the stage configuration is in a match and revise format.

    Examples:
        >>> is_match_revise(DictConfig({"_match_revise": []}))
        False
        >>> is_match_revise(DictConfig({"_match_revise": [{"_matcher": {"code": "CODE//TEMP"}}]}))
        True
        >>> is_match_revise(DictConfig({"foo": "bar"}))
        False
    """
    return bool(stage_cfg.get(MATCH_REVISE_KEY, False))


def validate_match_revise(stage_cfg: DictConfig):
    """Validate that the stage configuration is in a match and revise format.

    Examples:
        >>> validate_match_revise(DictConfig({"foo": []}))
        Traceback (most recent call last):
            ...
        ValueError: Stage configuration must contain a _match_revise key
        >>> validate_match_revise(DictConfig({"_match_revise": "foo"}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise options must be a list, got <class 'str'>
        >>> validate_match_revise(DictConfig({"_match_revise": [1]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 0 must be a dict, got <class 'int'>
        >>> validate_match_revise(DictConfig({"_match_revise": [{"_matcher": {"foo": "bar"}}, 1]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 1 must be a dict, got <class 'int'>
        >>> validate_match_revise(DictConfig({"_match_revise": [{"foo": "bar"}]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 0 must contain a _matcher key
        >>> validate_match_revise(DictConfig({"_match_revise": [{"_matcher": {32: "bar"}}]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 0 must contain a valid matcher in _matcher: ...
        >>> validate_match_revise(DictConfig({"_match_revise": [{"_matcher": {"code": "CODE//TEMP"}}]}))
    """

    if MATCH_REVISE_KEY not in stage_cfg:
        raise ValueError(f"Stage configuration must contain a {MATCH_REVISE_KEY} key")

    match_revise_options = stage_cfg[MATCH_REVISE_KEY]
    if not isinstance(match_revise_options, list | ListConfig):
        raise ValueError(f"Match revise options must be a list, got {type(match_revise_options)}")

    for i, match_revise_cfg in enumerate(match_revise_options):
        if not isinstance(match_revise_cfg, dict | DictConfig):
            raise ValueError(f"Match revise config {i} must be a dict, got {type(match_revise_cfg)}")

        if MATCHER_KEY not in match_revise_cfg:
            raise ValueError(f"Match revise config {i} must contain a {MATCHER_KEY} key")

        matcher_valid, matcher_errs = is_matcher(match_revise_cfg[MATCHER_KEY])
        if not matcher_valid:
            raise ValueError(
                f"Match revise config {i} must contain a valid matcher in {MATCHER_KEY}: {matcher_errs}"
            )


def match_revise_fntr(cfg: DictConfig, stage_cfg: DictConfig, compute_fn: ANY_COMPUTE_FN_T) -> COMPUTE_FN_T:
    """A functor that creates a match & revise compute function based on the given configuration.

    Stage configurations for match & revise must be in a match and revise format. Consider the below example,
    showing the ``stage_cfg`` object in ``yaml`` format:

        ```yaml
        global_arg_1: "foo"
        _match_revise:
          - _matcher: {code: "CODE//BAR"}
            local_arg_1: "bar"
          - _matcher: {code: "CODE//BAZ"}
            local_arg_1: "baz"
        ```

    This configuration will create a match & revise compute function that will filter the input DataFrame for
    rows that match the ``CODE//BAR`` code and apply the compute function with the ``local_arg_1=bar``
    parameter, and then filter the input DataFrame for rows that match the ``CODE//BAZ`` code and apply the
    compute function with the ``local_arg_1=baz`` parameter. Both of these local compute functions will be
    applied to the input DataFrame in sequence, and the resulting DataFrames will be concatenated alongside
    any of the dataframe that matches no matcher (which will be left unmodified) and merged in a sorted way
    that respects the ``subject_id``, ``time`` ordering first, then the order of the match & revise blocks
    themselves, then the order of the rows in each match & revise block output. Each local compute function
    will also use the ``global_arg_1=foo`` parameter.

    Args:
        cfg: The DictConfig configuration object.
        stage_cfg: The DictConfig stage configuration object. This stage configuration must be in a match and
            revise format, meaning it must have a key ``"_match_revise"`` that contains a list of local match
            & revise configurations. Each local match & revise configuration must contain a key ``"_matcher"``
            which links to the matcher configuration to use to filter the input DataFrame for the local
            compute execution, and all other keys are local configuration parameters to be used in the local
            compute execution.
        compute_fn: The compute function to bind to the match & revise configuration local arguments.

    Returns:
        A function that applies the match & revise compute function to the input DataFrame.

    Raises:
        ValueError: If the stage configuration is not in a match and revise format.

    Examples:
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 2, 2, 2],
        ...         "time": [1, 2, 2, 1, 1, 2],
        ...         "initial_idx": [0, 1, 2, 3, 4, 5],
        ...         "code": [
        ...             "FINAL",
        ...             "CODE//TEMP_2",
        ...             "CODE//TEMP_1",
        ...             "FINAL",
        ...             "CODE//TEMP_2",
        ...             "CODE//TEMP_1",
        ...         ],
        ...     }
        ... )
        >>> def compute_fn(df: pl.DataFrame, stage_cfg: DictConfig) -> pl.DataFrame:
        ...     return df.with_columns(
        ...         pl.col("code").str.slice(0, len("CODE//"))
        ...         + stage_cfg.local_code_mid
        ...         + "//"
        ...         + stage_cfg.global_code_end
        ...     )
        >>> stage_cfg = DictConfig(
        ...     {
        ...         "global_code_end": "foo",
        ...         "_match_revise": [
        ...             {"_matcher": {"code": "CODE//TEMP_1"}, "local_code_mid": "bar"},
        ...             {"_matcher": {"code": "CODE//TEMP_2"}, "local_code_mid": "baz"},
        ...         ],
        ...     }
        ... )
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        >>> match_revise_fn(df.lazy()).collect()
        shape: (6, 4)
        ┌────────────┬──────┬─────────────┬────────────────┐
        │ subject_id ┆ time ┆ initial_idx ┆ code           │
        │ ---        ┆ ---  ┆ ---         ┆ ---            │
        │ i64        ┆ i64  ┆ i64         ┆ str            │
        ╞════════════╪══════╪═════════════╪════════════════╡
        │ 1          ┆ 1    ┆ 0           ┆ FINAL          │
        │ 1          ┆ 2    ┆ 2           ┆ CODE//bar//foo │
        │ 1          ┆ 2    ┆ 1           ┆ CODE//baz//foo │
        │ 2          ┆ 1    ┆ 4           ┆ CODE//baz//foo │
        │ 2          ┆ 1    ┆ 3           ┆ FINAL          │
        │ 2          ┆ 2    ┆ 5           ┆ CODE//bar//foo │
        └────────────┴──────┴─────────────┴────────────────┘
        >>> stage_cfg = DictConfig(
        ...     {
        ...         "global_code_end": "foo",
        ...         "_match_revise": [
        ...             {"_matcher": {"code": "CODE//TEMP_2"}, "local_code_mid": "bizz"},
        ...             {
        ...                 "_matcher": {"code": "CODE//TEMP_1"},
        ...                 "local_code_mid": "foo",
        ...                 "global_code_end": "bar",
        ...             },
        ...         ],
        ...     }
        ... )
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        >>> match_revise_fn(df.lazy()).collect()
        shape: (6, 4)
        ┌────────────┬──────┬─────────────┬─────────────────┐
        │ subject_id ┆ time ┆ initial_idx ┆ code            │
        │ ---        ┆ ---  ┆ ---         ┆ ---             │
        │ i64        ┆ i64  ┆ i64         ┆ str             │
        ╞════════════╪══════╪═════════════╪═════════════════╡
        │ 1          ┆ 1    ┆ 0           ┆ FINAL           │
        │ 1          ┆ 2    ┆ 1           ┆ CODE//bizz//foo │
        │ 1          ┆ 2    ┆ 2           ┆ CODE//foo//bar  │
        │ 2          ┆ 1    ┆ 4           ┆ CODE//bizz//foo │
        │ 2          ┆ 1    ┆ 3           ┆ FINAL           │
        │ 2          ┆ 2    ┆ 5           ┆ CODE//foo//bar  │
        └────────────┴──────┴─────────────┴─────────────────┘
        >>> stage_cfg = DictConfig(
        ...     {"global_code_end": "foo", "_match_revise": [{"_matcher": {"missing": "CODE//TEMP_2"}}]}
        ... )
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        >>> match_revise_fn(df.lazy()).collect()
        Traceback (most recent call last):
            ...
        ValueError: Missing needed columns {'missing'} for local matcher 0:
            [(col("missing")) == ("CODE//TEMP_2")].all_horizontal()
        Columns available: 'code', 'initial_idx', 'subject_id', 'time'

        It will throw an error if the match and revise configuration is missing.
        >>> stage_cfg = DictConfig({"global_code_end": "foo"})
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        Traceback (most recent call last):
            ...
        ValueError: Invalid match and revise configuration...

        It does not accept invalid modes.
        >>> stage_cfg = DictConfig(
        ...     {
        ...         "global_code_end": "foo",
        ...         "_match_revise_mode": "foobar",
        ...         "_match_revise": [{"_matcher": {"code": "CODE//TEMP_2"}}],
        ...     }
        ... )
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        Traceback (most recent call last):
            ...
        ValueError: Invalid match and revise mode: foobar
    """
    try:
        validate_match_revise(stage_cfg)
    except ValueError as e:
        raise ValueError("Invalid match and revise configuration") from e

    stage_cfg = hydra.utils.instantiate(stage_cfg)

    match_revise_mode = stage_cfg.pop(MATCH_REVISE_MODE_KEY, "match_and_revise")
    if match_revise_mode not in {x.value for x in MatchReviseMode}:
        raise ValueError(f"Invalid match and revise mode: {match_revise_mode}")

    matchers_and_fns = []
    for match_revise_cfg in stage_cfg.pop(MATCH_REVISE_KEY):
        matcher, cols = matcher_to_expr(match_revise_cfg.pop(MATCHER_KEY))
        local_stage_cfg = DictConfig({**stage_cfg, **match_revise_cfg})
        local_compute_fn = bind_compute_fn(cfg, local_stage_cfg, compute_fn)

        matchers_and_fns.append((matcher, cols, local_compute_fn))

    @wraps(compute_fn)
    def match_revise_fn(df: DF_T) -> DF_T:
        matchable_df = df
        cols = set(df.collect_schema().names())

        revision_parts = []
        final_part_filters = []
        for i, (matcher_expr, need_cols, local_compute_fn) in enumerate(matchers_and_fns):
            if not need_cols.issubset(cols):
                cols_str = "', '".join(x for x in sorted(cols))
                raise ValueError(
                    f"Missing needed columns {need_cols - cols} for local matcher {i}: "
                    f"{matcher_expr}\nColumns available: '{cols_str}'"
                )
            matched_df = matchable_df.filter(matcher_expr)

            match match_revise_mode:
                case MatchReviseMode.MATCH_AND_REVISE:
                    matchable_df = matchable_df.filter(~matcher_expr)
                case MatchReviseMode.MULTI_MATCH_AND_REVISE:
                    final_part_filters.append(~matcher_expr)

            revision_parts.append(local_compute_fn(matched_df))

        if final_part_filters:
            revision_parts.append(matchable_df.filter(pl.all_horizontal(final_part_filters)))
        else:
            revision_parts.append(matchable_df)
        return pl.concat(revision_parts, how="vertical").sort(
            [DataSchema.subject_id_name, DataSchema.time_name], maintain_order=True
        )

    return match_revise_fn



"""Test set-up and fixtures code."""

import json
import tempfile
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from typing import Any
from unittest.mock import MagicMock, patch

import polars as pl
import pytest
from omegaconf import DictConfig


@contextmanager
def print_warnings(caplog: pytest.LogCaptureFixture):
    """Captures all logged warnings within this context block and prints them upon exit.

    This is useful in doctests, where you want to show printed outputs for documentation and testing purposes.
    """

    n_current_records = len(caplog.records)

    with caplog.at_level("WARNING"):
        yield
    # Print all captured warnings upon exit
    for record in caplog.records[n_current_records:]:
        print(f"Warning: {record.getMessage()}")


@pytest.fixture(autouse=True)
def __MEDS_transforms_setup_doctest_namespace(
    doctest_namespace: dict[str, Any],
    caplog: pytest.LogCaptureFixture,
    simple_static_MEDS,
) -> None:
    doctest_namespace.update(
        {
            "DictConfig": DictConfig,
            "MagicMock": MagicMock,
            "patch": patch,
            "simple_static_MEDS": simple_static_MEDS,
            "print_warnings": partial(print_warnings, caplog),
            "json": json,
            "pl": pl,
            "datetime": datetime,
            "tempfile": tempfile,
        }
    )



"""Functionality for reading input data shards."""

import logging
from collections.abc import Callable
from pathlib import Path

import polars as pl

from .types import DF_T

logger = logging.getLogger(__name__)

READ_FN_T = Callable[[Path], DF_T]


def read_df(in_fp: Path) -> DF_T:
    """A generic helper to read a dataframe without accounting for globs."""

    return pl.scan_parquet(in_fp, glob=False)


def read_and_filter_fntr(filter_expr: pl.Expr, read_fn: READ_FN_T = read_df) -> READ_FN_T:
    """Create a function that reads a DataFrame from a file and filters it based on a given expression.

    This is specified as a functor in this way to allow it to modify arbitrary other read functions for use in
    different mapreduce pipelines.

    Args:
        filter_expr: The filter expression to apply to the DataFrame.
        read_fn: The read function to use to read the DataFrame.

    Returns:
        A function that reads a DataFrame from a file and filters it based on the given expression.

    Examples:
        >>> dfs = {
        ...     "df1": pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
        ...     "df2": pl.DataFrame({"a": [4, 5, 6], "b": [7, 8, 9]}),
        ... }
        >>> read_fn = lambda key: dfs[key]
        >>> fn = read_and_filter_fntr((pl.col("a") % 2) == 0, read_fn)
        >>> fn("df1")
        shape: (1, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 2   ┆ 5   │
        └─────┴─────┘
        >>> fn("df2")
        shape: (2, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 4   ┆ 7   │
        │ 6   ┆ 9   │
        └─────┴─────┘
        >>> fn = read_and_filter_fntr((pl.col("b") % 2) == 0, read_fn)
        >>> fn("df1")
        shape: (2, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> fn("df2")
        shape: (1, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 5   ┆ 8   │
        └─────┴─────┘
    """

    def read_and_filter(in_fp: Path) -> DF_T:
        return read_fn(in_fp).filter(filter_expr)

    return read_and_filter



"""This file will use the automated stage example definitions to test each defined stage in this package."""

from MEDS_transforms.stages import StageExample


def test_stage_scenario(stage_example: StageExample):
    stage_example.test()



"""Generate Markdown documentation for registered stages.

This helper builds a ``StageDoc`` for each registered stage so it can be
rendered in the documentation site.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from .discovery import get_all_registered_stages


@dataclass
class StageDoc:
    """Markdown documentation page for a stage.

    Attributes:
        stage_name: Name of the stage.
        path: Relative path of the Markdown file.
        content: Markdown contents to write.
        edit_path: Optional repository path for MkDocs edit links.
    """

    stage_name: str
    path: Path
    content: str
    edit_path: Path | None = None


def generate_stage_docs(package: str, root: Path | None = None) -> list[StageDoc]:
    """Build documentation pages for all registered stages.

    Args:
        package: Package prefix used to filter stages from the registry.
        root: Repository root for computing edit links. Defaults to two
            directories above this file.

    Returns:
        A list of :class:`StageDoc` objects describing each page.

    Examples:
        >>> import textwrap
        >>> docs = generate_stage_docs("MEDS_transforms")
        >>> print(docs[0].stage_name)
        add_time_derived_measurements
        >>> print(docs[0].path)
        add_time_derived_measurements
        >>> print(textwrap.shorten(docs[0].content, width=50))
        # `add_time_derived_measurements` ## [...]

    It can also attempt to make docs for other packages; in this case, it will return None as there are no
    stages in this fake package:

        >>> generate_stage_docs("fake_package")
        []
    """

    root = Path(root) if root else Path(__file__).resolve().parents[1]

    docs: list[StageDoc] = []

    for stage_name, entry_point in sorted(get_all_registered_stages().items()):
        if not entry_point.module.startswith(package):
            continue

        stage = entry_point.load()
        doc_path = stage_name

        lines = [f"# `{stage_name}`"]

        if stage.test_cases:
            lines.append("")
            lines.append("## Examples")
            for scenario, example in stage.test_cases.items():
                scenario_name = scenario or "default"
                lines.extend(["", f"### {scenario_name}", "```", str(example), "```"])

        edit_path = stage.stage_dir.relative_to(root) if stage.stage_dir else None

        docs.append(StageDoc(stage_name, doc_path, "\n".join(lines), edit_path))

    return docs


__all__ = ["StageDoc", "generate_stage_docs"]



"""A functor to compute age measurements on a MEDS dataset."""

import logging
from collections.abc import Callable

import polars as pl
from meds import DataSchema
from omegaconf import DictConfig

from .utils import normalize_time_unit, unique_events

logger = logging.getLogger(__name__)


def age_fntr(cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
    """Create a function that adds a subject's age to a DataFrame.

    Args:
        cfg: The configuration for the age function. This must contain the following mandatory keys:
            - "DOB_code": The code for the date of birth event in the raw data. A string literal. Defaults to
              None. This is used preferentially over the "DOB_regex" key if both are provided.
            - "DOB_regex": A regex pattern to match the date of birth event in the raw data. Defaults to
              "^MEDS_BIRTH(//.*)?"
            - "age_code": The code for the age event in the output data. Defaults to "AGE".
            - "age_unit": The unit for the age event when converted to a numeric value in the output data.
              Defaults to "years"

    Returns:
        A function that returns the to-be-added "age" events with the subject's age for all input events with
        unique, non-null times in the data, for all subjects who have an observed date of birth. It does
        not add an event for times that are equal to the date of birth.

    Raises:
        ValueError: If the input unit is not recognized.

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 1, 1, 2, 2, 3, 3],
        ...         "time": [
        ...             None,
        ...             datetime(1990, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(2021, 1, 2),
        ...             datetime(1988, 1, 2),
        ...             datetime(2023, 1, 3),
        ...             datetime(2022, 1, 1),
        ...             datetime(2022, 1, 1),
        ...         ],
        ...         "code": ["static", "DOB", "lab//A", "lab//B", "rx", "DOB", "lab//A", "lab//B", "dx//1"],
        ...     },
        ...     schema={"subject_id": pl.UInt32, "time": pl.Datetime, "code": pl.Utf8},
        ... )
        >>> df
        shape: (9, 3)
        ┌────────────┬─────────────────────┬────────┐
        │ subject_id ┆ time                ┆ code   │
        │ ---        ┆ ---                 ┆ ---    │
        │ u32        ┆ datetime[μs]        ┆ str    │
        ╞════════════╪═════════════════════╪════════╡
        │ 1          ┆ null                ┆ static │
        │ 1          ┆ 1990-01-01 00:00:00 ┆ DOB    │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//A │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//B │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ rx     │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ DOB    │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ lab//A │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ lab//B │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ dx//1  │
        └────────────┴─────────────────────┴────────┘
        >>> age_fn = age_fntr(DictConfig({"DOB_code": "DOB", "age_code": "AGE", "age_unit": "years"}))
        >>> age_fn(df)
        shape: (3, 4)
        ┌────────────┬─────────────────────┬──────┬───────────────┐
        │ subject_id ┆ time                ┆ code ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---  ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str  ┆ f32           │
        ╞════════════╪═════════════════════╪══════╪═══════════════╡
        │ 1          ┆ 2021-01-01 00:00:00 ┆ AGE  ┆ 31.001347     │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ AGE  ┆ 31.004084     │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ AGE  ┆ 35.004169     │
        └────────────┴─────────────────────┴──────┴───────────────┘

    You can also use regular expressions for DOB codes. The default is "^MEDS_BIRTH(//.*)?", which matches
    optional prefixes in case the birth events have code modifiers:

        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 1, 1, 2, 2, 3, 3],
        ...         "time": [
        ...             None,
        ...             datetime(1990, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(2021, 1, 2),
        ...             datetime(1988, 1, 2),
        ...             datetime(2023, 1, 3),
        ...             datetime(2022, 1, 1),
        ...             datetime(2022, 1, 1),
        ...         ],
        ...         "code": [
        ...             "static",
        ...             "MEDS_BIRTH",
        ...             "lab//A",
        ...             "lab//B",
        ...             "rx",
        ...             "MEDS_BIRTH//home",
        ...             "lab//A",
        ...             "lab//B",
        ...             "dx//1",
        ...         ],
        ...     },
        ...     schema={"subject_id": pl.UInt32, "time": pl.Datetime, "code": pl.Utf8},
        ... )
        >>> cfg = DictConfig({"DOB_regex": "^MEDS_BIRTH(//.*)?", "age_code": "AGE", "age_unit": "days"})
        >>> age_fn = age_fntr(cfg)
        >>> age_fn(df)
        shape: (3, 4)
        ┌────────────┬─────────────────────┬──────┬───────────────┐
        │ subject_id ┆ time                ┆ code ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---  ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str  ┆ f32           │
        ╞════════════╪═════════════════════╪══════╪═══════════════╡
        │ 1          ┆ 2021-01-01 00:00:00 ┆ AGE  ┆ 11323.0       │
        │ 1          ┆ 2021-01-02 00:00:00 ┆ AGE  ┆ 11324.0       │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ AGE  ┆ 12785.0       │
        └────────────┴─────────────────────┴──────┴───────────────┘

    Errors are thrown if invalid units are provided or if the configuration doesn't have one of either
    `DOB_code` or `DOB_regex`:

        >>> age_fntr(DictConfig({"DOB_code": "DOB", "age_code": "AGE", "age_unit": "scores"}))
        Traceback (most recent call last):
            ...
        ValueError: Unknown time unit 'scores'. Valid units include:
        ...
        >>> age_fntr(DictConfig({"age_code": "AGE", "age_unit": "y"}))
        Traceback (most recent call last):
            ...
        ValueError: Either 'DOB_regex' or 'DOB_code' must be provided in the configuration.
    """

    canonical_unit, seconds_in_unit = normalize_time_unit(cfg.get("age_unit", "y"))
    microseconds_in_unit = int(1e6) * seconds_in_unit

    if cfg.get("DOB_code", None) is not None:
        is_dob = pl.col(DataSchema.code_name).str.contains(cfg.DOB_code, literal=True)
    elif cfg.get("DOB_regex", None) is not None:
        is_dob = pl.col(DataSchema.code_name).str.contains(cfg.DOB_regex)
    else:
        raise ValueError("Either 'DOB_regex' or 'DOB_code' must be provided in the configuration.")

    dob_col = "__dob"

    age_expr = (pl.col(DataSchema.time_name) - pl.col(dob_col)).dt.total_microseconds() / microseconds_in_unit
    age_expr = age_expr.cast(pl.Float32, strict=False)

    def fn(df: pl.LazyFrame) -> pl.LazyFrame:
        code_dtype = df.collect_schema().get(DataSchema.code_name, pl.Utf8)

        dobs = df.filter(is_dob).select(
            DataSchema.subject_id_name, pl.col(DataSchema.time_name).alias(dob_col)
        )
        events = unique_events(df)

        return (
            events.join(dobs, on=DataSchema.subject_id_name, how="inner", maintain_order="left")
            .select(
                DataSchema.subject_id_name,
                DataSchema.time_name,
                pl.lit(cfg.get("age_code", "AGE"), dtype=code_dtype).alias(DataSchema.code_name),
                age_expr.alias(DataSchema.numeric_value_name),
            )
            .drop_nulls(subset=[DataSchema.numeric_value_name])
            .filter(pl.col(DataSchema.numeric_value_name) > 0)
        )

    return fn



"""Simple helper functions to define a consistent code vocabulary for normalizing a MEDS dataset."""

import logging
from collections.abc import Callable
from enum import StrEnum
from pathlib import Path

import polars as pl
from omegaconf import DictConfig, OmegaConf

from .. import Stage

logger = logging.getLogger(__name__)


class VocabularyOrdering(StrEnum):
    """Enumeration of different ways a vocabulary order can be selected.

    These are stored as a `StrEnum` so that they can be easily specified by the user in a configuration file
    or on the command line.

    Currently, only one ordering method is supported, but others can be added, such as a frequency-based
    ordering so that the most frequent codes have the smallest indices.

    Args:
        "lexicographic": Assigns vocabulary indices to codes and code modifiers via a lexicographic order.
    """

    LEXICOGRAPHIC = "lexicographic"


INDEX_ASSIGNMENT_FN = Callable[[pl.DataFrame, list[str]], pl.DataFrame]


def validate_code_metadata(code_metadata: pl.DataFrame, code_modifiers: list[str]):
    """Validate the code metadata has the requisite columns and is unique.

    Args:
        code_metadata: Metadata about the codes in the MEDS dataset, with a column `code` and a collection
            of code modifier columns.
        code_modifiers: The names of the code modifier columns in the `code_metadata` dataset.

    Raises:
        KeyError: If the `code_metadata` dataset does not contain the specified `code_modifiers` or `code`
            columns.
        ValueError: If the `code_metadata` dataset is not unique on the `code` and `code_modifiers` columns.

    Examples:
        >>> code_metadata = pl.DataFrame(
        ...     {
        ...         "code": ["A", "B", "A", "A"],
        ...         "modifier1": ["X", "D", "Z", "Z"],
        ...         "modifier2": [None, None, None, 3],
        ...     }
        ... )
        >>> validate_code_metadata(code_metadata, ["modifier1", "modifier2"])
        >>> # This returns None in the absence of an exception.
        >>> code_metadata = pl.DataFrame(
        ...     {
        ...         "code": ["A", "B", "A", "A"],
        ...         "modifier1": ["X", "D", "Z", "Z"],
        ...         "modifier2": [None, None, None, 3],
        ...     }
        ... )
        >>> validate_code_metadata(code_metadata, ["modifier1", "modifier2", "missing_modifier"])
        Traceback (most recent call last):
            ...
        KeyError: "The following columns are not present in the code metadata: 'missing_modifier'."
        >>> code_metadata = pl.DataFrame(
        ...     {
        ...         "code": ["A", "B", "A", "A", "B", "B"],
        ...         "modifier1": ["X", "D", "Z", "Z", "Y", "Y"],
        ...         "modifier2": [None, None, None, None, 2, 1],
        ...     }
        ... )
        >>> validate_code_metadata(code_metadata, ["modifier1", "modifier2"])
        Traceback (most recent call last):
            ...
        ValueError: The code and code modifiers are not unique:
        shape: (1, 4)
        ┌──────┬───────────┬───────────┬───────┐
        │ code ┆ modifier1 ┆ modifier2 ┆ count │
        │ ---  ┆ ---       ┆ ---       ┆ ---   │
        │ str  ┆ str       ┆ i64       ┆ u32   │
        ╞══════╪═══════════╪═══════════╪═══════╡
        │ A    ┆ Z         ┆ null      ┆ 2     │
        └──────┴───────────┴───────────┴───────┘
    """

    cols = ["code", *code_modifiers]

    # Check that the code and code modifiers are present in the code metadata
    if not set(cols).issubset(code_metadata.columns):
        missing_cols = set(cols) - set(code_metadata.columns)
        missing_cols_str = "', '".join(missing_cols)
        raise KeyError(f"The following columns are not present in the code metadata: '{missing_cols_str}'.")

    # Check that the code and code modifiers are unique
    n_unique_codes = code_metadata.n_unique(cols)
    n_total_rows = len(code_metadata)

    if n_unique_codes != n_total_rows:
        code_counts = code_metadata.group_by(cols).agg(pl.len().alias("count")).sort("count", descending=True)
        extra_codes = code_counts.filter(pl.col("count") > 1)
        raise ValueError(f"The code and code modifiers are not unique:\n{extra_codes.head(100)}")


def lexicographic_indices(code_metadata: pl.DataFrame, code_modifiers: list[str]) -> pl.DataFrame:
    """Assign vocabulary indices to codes and code modifiers via a lexicographic order.

    Args:
        code_metadata: Metadata about the codes in the MEDS dataset, with a column `code` and a collection
            of code modifier columns.
        code_modifiers: The names of the code modifier columns in the `code_metadata` dataset. Each of these
            columns should be lexicographically orderable.

    Returns:
        The code metadata dataframe with an additional column added that hasvocabulary token indices to the
        code + modifier unique combinations in the `code_metadata` dataset. The expression will be aliased to
        "code/vocab_index", and will be of the smallest unsigned dtype possible given the number of included
        vocabulary elements. The given vocabulary indices will correspond to the following order:
          - The index `0` will be assigned to a sentinel, `"UNK"` code for codes/modifiers not present in the
            vocabulary.
          - The remaining indices will be assigned to the unique code + modifier combinations such that
            sorting the code + modifier combinations by the assigned index will result in a lexicographically
            ordered list of codes + modifiers, sorting first by code and subsequently by each modifier column,
            in order of specification. The sort will go from smallest lexiographic value to largest (e.g., be
            an ascending sort). `null` values in the modifier columns (`null`s are disallowed in the code
            columns) will be treated as the smallest possible value in the lexicographic order.

    Raises:
        KeyError: If the `code_metadata` dataset does not contain the specified `code_modifiers` or `code`
            columns.
        ValueError: If the `code_metadata` dataset is not unique on the `code` and `code_modifiers` columns.
        ValueError: If the `code` and `code_modifier` columns are not all lexicographically orderable.

    Examples:
        >>> code_metadata = pl.DataFrame(
        ...     {
        ...         "code": ["A", "B", "A", "A", "B", "B"],
        ...         "modifier1": ["X", "D", None, "Z", "Y", "Y"],
        ...         "modifier2": [None, None, None, None, 2, 1],
        ...     }
        ... )
        >>> code_modifiers = ["modifier1", "modifier2"]
        >>> lexicographic_indices(code_metadata, code_modifiers)
        shape: (6, 4)
        ┌──────┬───────────┬───────────┬──────────────────┐
        │ code ┆ modifier1 ┆ modifier2 ┆ code/vocab_index │
        │ ---  ┆ ---       ┆ ---       ┆ ---              │
        │ str  ┆ str       ┆ i64       ┆ u8               │
        ╞══════╪═══════════╪═══════════╪══════════════════╡
        │ A    ┆ X         ┆ null      ┆ 2                │
        │ B    ┆ D         ┆ null      ┆ 4                │
        │ A    ┆ null      ┆ null      ┆ 1                │
        │ A    ┆ Z         ┆ null      ┆ 3                │
        │ B    ┆ Y         ┆ 2         ┆ 6                │
        │ B    ┆ Y         ┆ 1         ┆ 5                │
        └──────┴───────────┴───────────┴──────────────────┘
    """

    validate_code_metadata(code_metadata, code_modifiers)

    # We'll perform this sort in three steps. To guide this, consider as an example that our set of codes is
    # ["B", "D", "A", "C"]. For this, we want to produce a set of indices that correspond to the lexicographic
    # order of the codes, starting at 1; namely, [2, 4, 1, 3]. This is because "B" is the 2nd letter, "D" the
    # fourth, and so on.
    # 1. First, we'll use an `pl.arg_sort_by` to produce a set of indices that, were we to select the rows in
    #    that order, would give us the codes in lexicographic order. This is _not_ the final order -- it tells
    #    us what row we'd need to put in each position to _get_ the codes in sorted order.
    #    E.g., if our codes were ["B", "D", "A", "C"], the result of this step would be [2, 0, 3, 1], because
    #    we'd need to get the 2nd row to have the first lexicographically ordered code, the 0th row to have
    #    the second, and so on.
    # 2. Second, we'll use _another_ `pl.arg_sort_by` to identify the row indices that would sort the very
    #    sort indices we just produced. This works because the index of the destination each row would have in
    #    the final sorted array is exactly the position that that row's index appears in the sort indices
    #    produced in step 1, by definition. And, in the second arg-sort, when we ask "which row do we need to
    #    grab to fill slot $j$ in this array with the sorted element that would belong at position $j$ of this
    #    set of numbers between $0$ and $N-1$, we are really asking which row has $j$ in it now, which is
    #    exactly the lexicographically ordered index.
    # 3. Finally, third, we will add one (to start at one) and shrink the dtype.
    #
    # Note that we use this algorithm over something like just sorting the whole dataframe once then assigning
    # integer indices is that this approach merely assigns indices, and does not change the order of the
    # dataframe, and similarly does not require actually touching any of the memory of the dataframe. Though,
    # admittedly, it is not clear how significant this choice is in practice.

    sort_cols = ["code", *code_modifiers]

    return code_metadata.with_columns(
        (pl.arg_sort_by(pl.arg_sort_by(sort_cols, descending=False, nulls_last=False)) + 1)
        .shrink_dtype()
        .alias("code/vocab_index")
    )


VocabularyOrdering_METHODS: dict[VocabularyOrdering, INDEX_ASSIGNMENT_FN] = {
    VocabularyOrdering.LEXICOGRAPHIC: lexicographic_indices,
}

VOCABULARY_SCHEMA_UPDATES = {"code/vocab_index": pl.UInt8}


@Stage.register(output_schema_updates=VOCABULARY_SCHEMA_UPDATES, is_metadata=True)
def main(cfg: DictConfig):
    """Assigns integral vocabulary IDs to codes in the metadata file, for use in tokenizing the dataset.

    This stage merely modifies the specified input code metadata file, by assigning a `code/vocab_index`
    column to the input dataframe whose entries correspond to the ordered vocabulary index realized by the
    specified ordering method over all unique codes.

    Args:
        stage_configs.fit_vocabulary_indices.ordering_method: How the code vocabulary should be ordered when
            vocab indices are assigned. Currently, only `lexographic` is supported. File a GitHub issue if
            you'd like other ordering methods supported.

    Raises:
        ValueError: If the ordering method is not one of the supported methods.

    Examples:

        To show this in action, we'll use this example code metadata file:

        >>> code_metadata = pl.DataFrame(
        ...     {
        ...         "code": ["A", "B", "A", "A", "B", "B"],
        ...         "modifier1": ["X", "D", None, "Z", "Y", "Y"],
        ...         "modifier2": [None, None, None, None, 2, 1],
        ...     }
        ... )

        We'll also use the following global configuration file:

        >>> cfg = DictConfig(
        ...     {
        ...         "stage": "fit_vocabulary_indices",
        ...         "code_modifier_columns": ["modifier1", "modifier2"],
        ...         "stage_cfg": {
        ...             "metadata_input_dir": "???",  # Will be assigned later in the test.
        ...             "reducer_output_dir": "???",  # Will be assigned later in the test.
        ...             "ordering_method": "lexicographic",
        ...         },
        ...     }
        ... )

        Now, we'll run the stage with the above code metadata and configuration:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     # Set up the input and output directories.
        ...     metadata_input_dir = Path(tmpdir) / "input"
        ...     metadata_input_dir.mkdir(parents=True, exist_ok=True)
        ...     code_metadata.write_parquet(metadata_input_dir / "codes.parquet")
        ...     output_dir = Path(tmpdir) / "output"
        ...     cfg.stage_cfg.metadata_input_dir = str(metadata_input_dir)
        ...     cfg.stage_cfg.reducer_output_dir = str(output_dir)
        ...     # Run the stage
        ...     main(cfg)
        ...     # Read the output file and print it.
        ...     pl.read_parquet(output_dir / "codes.parquet")
        shape: (6, 4)
        ┌──────┬───────────┬───────────┬──────────────────┐
        │ code ┆ modifier1 ┆ modifier2 ┆ code/vocab_index │
        │ ---  ┆ ---       ┆ ---       ┆ ---              │
        │ str  ┆ str       ┆ i64       ┆ u8               │
        ╞══════╪═══════════╪═══════════╪══════════════════╡
        │ A    ┆ X         ┆ null      ┆ 2                │
        │ B    ┆ D         ┆ null      ┆ 4                │
        │ A    ┆ null      ┆ null      ┆ 1                │
        │ A    ┆ Z         ┆ null      ┆ 3                │
        │ B    ┆ Y         ┆ 2         ┆ 6                │
        │ B    ┆ Y         ┆ 1         ┆ 5                │
        └──────┴───────────┴───────────┴──────────────────┘

        If the `ordering_method` were set to another parameter, an error would be thrown (as right now only
        lexicographic ordering is supported):

        >>> cfg.stage_cfg.ordering_method = "other"
        >>> main(cfg)
        Traceback (most recent call last):
            ...
        ValueError: Invalid ordering method: other. Expected one of lexicographic

        Note that if the code and code modifiers in the metadata file are not unique, an error will be thrown
        as well -- this is because without unique assignments between codes/modifiers and indices, you would
        not be able to convert between codes and indices in the raw data. We can see that in action here by
        removing the (necessary) code modifiers from our configuration file:

        >>> cfg.stage_cfg.ordering_method = "lexicographic"
        >>> cfg.code_modifier_columns = None
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     # Set up the input and output directories.
        ...     metadata_input_dir = Path(tmpdir) / "input"
        ...     metadata_input_dir.mkdir(parents=True, exist_ok=True)
        ...     code_metadata.write_parquet(metadata_input_dir / "codes.parquet")
        ...     output_dir = Path(tmpdir) / "output"
        ...     cfg.stage_cfg.metadata_input_dir = str(metadata_input_dir)
        ...     cfg.stage_cfg.reducer_output_dir = str(output_dir)
        ...     # Run the stage
        ...     main(cfg)
        Traceback (most recent call last):
            ...
        ValueError: The code and code modifiers are not unique: ...
    """

    logger.info(
        f"Running with config:\n{OmegaConf.to_yaml(cfg)}\n"
        f"Stage: {cfg.stage}\n\n"
        f"Stage config:\n{OmegaConf.to_yaml(cfg.stage_cfg)}"
    )

    ordering_method = cfg.stage_cfg.get("ordering_method", VocabularyOrdering.LEXICOGRAPHIC)

    if ordering_method not in VocabularyOrdering_METHODS:
        raise ValueError(
            f"Invalid ordering method: {ordering_method}. "
            f"Expected one of {', '.join(VocabularyOrdering_METHODS.keys())}"
        )

    metadata_input_dir = Path(cfg.stage_cfg.metadata_input_dir)
    output_dir = Path(cfg.stage_cfg.reducer_output_dir)

    code_metadata = pl.read_parquet(metadata_input_dir / "codes.parquet", use_pyarrow=True)

    logger.info(f"Assigning code vocabulary indices via a {ordering_method} order.")
    ordering_fn = VocabularyOrdering_METHODS[ordering_method]

    code_modifiers = cfg.get("code_modifier_columns", None)
    if code_modifiers is None:
        code_modifiers = []

    code_metadata = ordering_fn(code_metadata, code_modifiers)

    output_fp = output_dir / "codes.parquet"
    output_fp.parent.mkdir(parents=True, exist_ok=True)
    logger.info(f"Indices assigned. Writing to {output_fp}")

    code_metadata.write_parquet(output_fp, use_pyarrow=True)

    logger.info(f"Done with {cfg.stage}")



from .read_fn import READ_FN_T, read_and_filter_fntr, read_df
from .types import DF_T
from .write_fn import WRITE_FN_T, write_df

__all__ = ["DF_T", "READ_FN_T", "WRITE_FN_T", "read_and_filter_fntr", "read_df", "write_df"]



import importlib
import subprocess
import sys
import tempfile
from io import StringIO
from pathlib import Path

import polars as pl
from polars.testing import assert_frame_equal
from pretty_print_directory import print_directory


def test_simple_example_pipeline():
    examples_dir = Path(__file__).resolve().parents[1] / "example"
    pkg_dir = examples_dir / "simple_example_pkg"
    subprocess.run([sys.executable, "-m", "pip", "install", "-e", str(pkg_dir)], check=True)

    sys.path.insert(0, str(pkg_dir / "src"))
    importlib.invalidate_caches()

    input_dir = examples_dir / "data"
    regression_target = examples_dir / "output_data"

    want_data_files = list((regression_target / "data").rglob("*.parquet"))
    if len(want_data_files) == 0:
        raise ValueError("No regression target data files found.")

    want_metadata_files = list((regression_target / "metadata").rglob("*.*"))
    if len(want_metadata_files) == 0:
        raise ValueError("No regression target metadata files found.")

    with tempfile.TemporaryDirectory() as tmpdir:
        output_dir = Path(tmpdir)

        cmd = [
            "MEDS_transform-pipeline",
            "pkg://simple_example_pkg.pipelines.example_pipeline.yaml",
            "--overrides",
            f"input_dir={input_dir!s}",
            f"output_dir={output_dir!s}",
        ]

        out = subprocess.run(cmd, check=False, capture_output=True)

        stdout = out.stdout.decode("utf-8")
        stderr = out.stderr.decode("utf-8")

        sio = StringIO()
        print_directory(output_dir, file=sio)
        dir_contents = sio.getvalue()

        pipeline_log_fp = output_dir / ".logs" / "pipeline.log"
        pipeline_log = pipeline_log_fp.read_text() if pipeline_log_fp.exists() else "DOES NOT EXIST"

        msg_parts = {
            "Directory contents": dir_contents,
            "Pipeline log": pipeline_log,
            "stdout": stdout,
            "stderr": stderr,
        }
        msg_parts_str = "\n".join(f"{k}:\n{v}" for k, v in msg_parts.items())

        assert out.returncode == 0, f"Error running pipeline:\n{msg_parts_str}"

        got_data_files = list((output_dir / "data").rglob("*.parquet"))
        assert len(got_data_files) == len(want_data_files), (
            f"Expected {len(want_data_files)} data files, but got {len(got_data_files)}.\n{msg_parts_str}"
        )

        for fp in got_data_files:
            fn = fp.relative_to(output_dir)
            want_fp = regression_target / fn
            assert want_fp.exists(), f"Expected output file {want_fp} does not exist."

            want_df = pl.read_parquet(want_fp)
            got_df = pl.read_parquet(fp)
            assert_frame_equal(want_df, got_df)

        got_metadata_files = list((output_dir / "metadata").rglob("*.*"))
        assert len(got_metadata_files) == len(want_metadata_files), (
            f"Expected {len(want_metadata_files)} metadata files, but got {len(got_metadata_files)}.\n"
            f"{msg_parts_str}"
        )

        for fp in got_metadata_files:
            fn = fp.relative_to(output_dir)
            want_fp = regression_target / fn
            assert want_fp.exists(), f"Expected metadata file {want_fp} does not exist."

            if fp.suffix == ".parquet":
                want_df = pl.read_parquet(want_fp)
                got_df = pl.read_parquet(fp)
                assert_frame_equal(want_df, got_df)
            elif fp.suffix == ".json":
                assert want_fp.read_text() == fp.read_text(), (
                    f"Expected metadata file {want_fp} does not match got {fp}."
                )



"""Tests a multi-stage pre-processing pipeline via the Runner utility. Only checks final outputs.

In this test, the following stages are run:
  - filter_subjects
  - add_time_derived_measurements
  - fit_outlier_detection
  - occlude_outliers
  - fit_normalization
  - fit_vocabulary_indices
  - normalization

The stage configuration arguments will be as given in the yaml block below:
"""

import subprocess
import tempfile
from pathlib import Path

import pytest

from MEDS_transforms.pytest_plugin import pipeline_tester

PIPELINE_YAML = """
input_dir: {input_dir}
output_dir: {output_dir}

description: "A test pipeline for the MEDS-transforms pipeline runner."

stages:
  - filter_subjects:
      min_events_per_subject: 5
  - add_time_derived_measurements:
      age:
        DOB_code: "DOB"
        age_code: "AGE"
        age_unit: "years"
      time_of_day:
        time_of_day_code: "TIME_OF_DAY"
        endpoints: [6, 12, 18, 24]
  - fit_outlier_detection:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - occlude_outliers:
      stddev_cutoff: 1
  - fit_normalization:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "code/n_occurrences"
        - "code/n_subjects"
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - fit_vocabulary_indices
  - normalization
"""

STAGE_SCENARIO_SEQUENCE = [
    "filter_subjects",
    "add_time_derived_measurements/in_example_pipeline",
    "aggregate_code_metadata/in_example_pipeline/fit_outlier_detection",
    "occlude_outliers/in_example_pipeline",
    "aggregate_code_metadata/in_example_pipeline/fit_normalization",
    "fit_vocabulary_indices/in_example_pipeline",
    "normalization/in_example_pipeline",
]


# Normally, you wouldn't need to specify all of these scripts, but in testing with local scripts we need to
# specify them all as they need to point to their python paths.
PARALLEL_STAGE_RUNNER_YAML = """
parallelize:
  n_workers: 2
  launcher: "joblib"
"""


def test_example_pipeline():
    pipeline_tester(PIPELINE_YAML, None, STAGE_SCENARIO_SEQUENCE)


@pytest.mark.parallelized
def test_example_pipeline_parallel_stage_runner():
    log_out = pipeline_tester(PIPELINE_YAML, PARALLEL_STAGE_RUNNER_YAML, STAGE_SCENARIO_SEQUENCE)

    assert "Parallelization configuration loaded from stage runner" in log_out, (
        f"Expected parallelization message not found in log output:\n{log_out}"
    )


PIPELINE_YAML_PARALLEL = """
input_dir: {input_dir}
output_dir: {output_dir}

description: "A test pipeline for the MEDS-transforms pipeline runner."

stages:
  - filter_subjects:
      min_events_per_subject: 5
  - add_time_derived_measurements:
      age:
        DOB_code: "DOB"
        age_code: "AGE"
        age_unit: "years"
      time_of_day:
        time_of_day_code: "TIME_OF_DAY"
        endpoints: [6, 12, 18, 24]
  - fit_outlier_detection:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - occlude_outliers:
      stddev_cutoff: 1
  - fit_normalization:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "code/n_occurrences"
        - "code/n_subjects"
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - fit_vocabulary_indices
  - normalization

parallelize:
  n_workers: 2
  launcher: "joblib"
"""


@pytest.mark.parallelized
def test_example_pipeline_parallel_pipeline_cfg():
    log_out = pipeline_tester(PIPELINE_YAML_PARALLEL, None, STAGE_SCENARIO_SEQUENCE)
    assert "Parallelization configuration loaded from pipeline config" in log_out, (
        f"Expected parallelization message not found in log output:\n{log_out}"
    )


def test_pipeline_help():
    out = subprocess.run("MEDS_transform-pipeline -h", shell=True, check=True, capture_output=True)
    help_text = out.stdout.decode("utf-8")
    assert "usage:" in help_text.lower()
    assert "pipeline_config_fp" in help_text


def test_pipeline_runner_with_done_file():
    """Test that the pipeline runner does nothing when a global done file is present at the start."""

    with tempfile.TemporaryDirectory() as temp_dir:
        root = Path(temp_dir)
        input_dir = root / "input"
        output_dir = root / "output"
        log_dir = output_dir / ".logs"

        global_done_file = log_dir / "_all_stages.done"
        global_done_file.parent.mkdir(parents=True, exist_ok=True)
        global_done_file.touch()

        pipeline_config = PIPELINE_YAML.format(input_dir=input_dir, output_dir=output_dir)

        pipeline_config_path = root / "pipeline_config.yaml"
        with open(pipeline_config_path, "w") as f:
            f.write(pipeline_config)

        # Run the pipeline
        out = subprocess.run(
            f"MEDS_transform-pipeline {pipeline_config_path!s}", shell=True, check=False, capture_output=True
        )

        stdout = out.stdout.decode("utf-8")
        stderr = out.stderr.decode("utf-8")

        assert out.returncode == 0, f"Error running pipeline:\n{stdout}\n{stderr}"

        want_txt = "All stages are already complete. Exiting."

        pipeline_log = log_dir / "pipeline.log"
        assert pipeline_log.exists(), "Pipeline log file does not exist."
        assert pipeline_log.is_file(), "Pipeline log is not a file."
        assert want_txt in pipeline_log.read_text(), "Pipeline log does not contain expected text."


PIPELINE_YAML_NO_OUTPUT = """
input_dir: {input_dir}
output_dir: ???

description: "A test pipeline for the MEDS-transforms pipeline runner."

stages:
  - filter_subjects:
      min_events_per_subject: 5
  - add_time_derived_measurements:
      age:
        DOB_code: "DOB"
        age_code: "AGE"
        age_unit: "years"
      time_of_day:
        time_of_day_code: "TIME_OF_DAY"
        endpoints: [6, 12, 18, 24]
  - fit_outlier_detection:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - occlude_outliers:
      stddev_cutoff: 1
  - fit_normalization:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "code/n_occurrences"
        - "code/n_subjects"
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - fit_vocabulary_indices
  - normalization
"""


def test_errors_without_output_dir():
    with tempfile.TemporaryDirectory() as temp_dir:
        root = Path(temp_dir)
        input_dir = root / "input"

        pipeline_config = PIPELINE_YAML_NO_OUTPUT.format(input_dir=input_dir)

        pipeline_config_path = root / "pipeline_config.yaml"
        with open(pipeline_config_path, "w") as f:
            f.write(pipeline_config)

        cmd = f"MEDS_transform-pipeline {pipeline_config_path!s}"

        # Run the pipeline
        out = subprocess.run(cmd, shell=True, check=False, capture_output=True)

        stdout = out.stdout.decode("utf-8")
        stderr = out.stderr.decode("utf-8")

        assert out.returncode == 1, f"Pipeline should error but didn't!\n{stdout}\n{stderr}"

        want_txt = "ValueError: Pipeline configuration or override must specify an 'output_dir'"

        assert want_txt in stderr, "Pipeline did not error as expected with missing output_dir."


def test_additional_pipeline_args():
    with tempfile.TemporaryDirectory() as temp_dir:
        root = Path(temp_dir)
        input_dir = root / "input"
        output_dir = root / "output"
        log_dir = output_dir / ".logs"

        global_done_file = log_dir / "_all_stages.done"
        global_done_file.parent.mkdir(parents=True, exist_ok=True)
        global_done_file.touch()

        pipeline_config = PIPELINE_YAML_NO_OUTPUT.format(input_dir=input_dir)

        pipeline_config_path = root / "pipeline_config.yaml"
        with open(pipeline_config_path, "w") as f:
            f.write(pipeline_config)

        cmd = f"MEDS_transform-pipeline {pipeline_config_path!s} --overrides output_dir={output_dir!s}"

        # Run the pipeline
        out = subprocess.run(cmd, shell=True, check=False, capture_output=True)

        stdout = out.stdout.decode("utf-8")
        stderr = out.stderr.decode("utf-8")

        assert out.returncode == 0, f"Error running pipeline:\n{stdout}\n{stderr}"

        want_txt = "All stages are already complete. Exiting."

        pipeline_log = log_dir / "pipeline.log"
        assert pipeline_log.exists(), "Pipeline log file does not exist."
        assert pipeline_log.is_file(), "Pipeline log is not a file."
        assert want_txt in pipeline_log.read_text(), "Pipeline log does not contain expected text."



"""A polars-to-polars transformation function for filtering subjects by sequence length."""

import logging
from collections.abc import Callable

import polars as pl
from omegaconf import DictConfig, ListConfig

from .. import Stage

logger = logging.getLogger(__name__)


def get_smallest_valid_uint_type(num: int | float | pl.Expr) -> pl.DataType:
    """Returns the smallest valid unsigned integral type for an ID variable with `num` unique options.

    Args:
        num: The number of IDs that must be uniquely expressed.

    Raises:
        ValueError: If there is no unsigned int type big enough to express the passed number of ID
            variables.

    Examples:
        >>> get_smallest_valid_uint_type(num=1)
        UInt8
        >>> get_smallest_valid_uint_type(num=2**8 - 1)
        UInt16
        >>> get_smallest_valid_uint_type(num=2**16 - 1)
        UInt32
        >>> get_smallest_valid_uint_type(num=2**32 - 1)
        UInt64
        >>> get_smallest_valid_uint_type(num=2**64 - 1)
        Traceback (most recent call last):
            ...
        ValueError: Value is too large to be expressed as an int!
    """
    if num >= (2**64) - 1:
        raise ValueError("Value is too large to be expressed as an int!")
    if num >= (2**32) - 1:
        return pl.UInt64
    elif num >= (2**16) - 1:
        return pl.UInt32
    elif num >= (2**8) - 1:
        return pl.UInt16
    else:
        return pl.UInt8


@Stage.register
def reorder_measurements(
    stage_cfg: DictConfig, code_metadata: pl.DataFrame, code_modifiers: list[str] | None = None
) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """Re-orders a dataframe within the temporal and subject ID ordering via a specified code order.

    Args:
        stage_cfg: The stage-specific configuration object which contains the `ordered_code_patterns` field
            that defines the order of the codes within each subject event (unique timepoint). Each element of
            this list should be a regex pattern that matches codes that should be re-ordered at the index of
            the regex pattern in the list. Codes are matched in the order of the list, and if a code matches
            multiple regex patterns, it will be ordered by the first regex pattern that matches it.
        code_metadata: The metadata DataFrame that contains the code column and any code modifiers. This must
            contain a `code` column that has all codes in the dataset.
        code_modifiers: The list of columns that are "code modifiers", meaning they should be used along with
            the code in join or group-by operations.

    Returns:
        A function with signature `Callable[[pl.LazyFrame], pl.LazyFrame]` that re-orders the input DataFrame.

    Examples:
        >>> code_metadata_df = pl.DataFrame({"code": ["A", "A", "B", "C"], "modifier1": [1, 2, 1, 2]})
        >>> data = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 2, 2],
        ...         "time": [1, 1, 1, 1],
        ...         "code": ["A", "B", "A", "C"],
        ...         "modifier1": [1, 2, 1, 2],
        ...     }
        ... )
        >>> stage_cfg = DictConfig({"ordered_code_patterns": ["B", "A"]})
        >>> fn = reorder_measurements(stage_cfg, code_metadata_df, ["modifier1"])
        >>> fn(data.lazy()).collect()
        shape: (4, 4)
        ┌────────────┬──────┬──────┬───────────┐
        │ subject_id ┆ time ┆ code ┆ modifier1 │
        │ ---        ┆ ---  ┆ ---  ┆ ---       │
        │ i64        ┆ i64  ┆ str  ┆ i64       │
        ╞════════════╪══════╪══════╪═══════════╡
        │ 1          ┆ 1    ┆ B    ┆ 2         │
        │ 1          ┆ 1    ┆ A    ┆ 1         │
        │ 2          ┆ 1    ┆ A    ┆ 1         │
        │ 2          ┆ 1    ┆ C    ┆ 2         │
        └────────────┴──────┴──────┴───────────┘
        >>> code_metadata_df = pl.DataFrame(
        ...     {
        ...         "code": ["LAB//foo", "ADMISSION//bar", "LAB//baz", "ADMISSION//qux", "DISCHARGE"],
        ...     }
        ... )
        >>> data = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 2, 2, 2],
        ...         "time": [1, 1, 1, 1, 2, 3],
        ...         "code": [
        ...             "LAB//foo",
        ...             "ADMISSION//bar",
        ...             "LAB//baz",
        ...             "ADMISSION//qux",
        ...             "DISCHARGE",
        ...             "LAB//baz",
        ...         ],
        ...     }
        ... )
        >>> stage_cfg = DictConfig(
        ...     {"ordered_code_patterns": ["ADMISSION.*", "LAB//baza", "LAB//f$", "LAB//b.*", "DISCHARGE"]}
        ... )
        >>> fn = reorder_measurements(stage_cfg, code_metadata_df)
        >>> fn(data.lazy()).collect()
        shape: (6, 3)
        ┌────────────┬──────┬────────────────┐
        │ subject_id ┆ time ┆ code           │
        │ ---        ┆ ---  ┆ ---            │
        │ i64        ┆ i64  ┆ str            │
        ╞════════════╪══════╪════════════════╡
        │ 1          ┆ 1    ┆ ADMISSION//bar │
        │ 1          ┆ 1    ┆ LAB//baz       │
        │ 1          ┆ 1    ┆ LAB//foo       │
        │ 2          ┆ 1    ┆ ADMISSION//qux │
        │ 2          ┆ 2    ┆ DISCHARGE      │
        │ 2          ┆ 3    ┆ LAB//baz       │
        └────────────┴──────┴────────────────┘
        >>> fn = reorder_measurements({}, code_metadata_df)
        >>> fn(data.lazy()).collect()
        shape: (6, 3)
        ┌────────────┬──────┬────────────────┐
        │ subject_id ┆ time ┆ code           │
        │ ---        ┆ ---  ┆ ---            │
        │ i64        ┆ i64  ┆ str            │
        ╞════════════╪══════╪════════════════╡
        │ 1          ┆ 1    ┆ LAB//foo       │
        │ 1          ┆ 1    ┆ ADMISSION//bar │
        │ 1          ┆ 1    ┆ LAB//baz       │
        │ 2          ┆ 1    ┆ ADMISSION//qux │
        │ 2          ┆ 2    ┆ DISCHARGE      │
        │ 2          ┆ 3    ┆ LAB//baz       │
        └────────────┴──────┴────────────────┘
        >>> fn = reorder_measurements({"ordered_code_patterns": "foo"}, code_metadata_df)
        Traceback (most recent call last):
            ...
        ValueError: The 'ordered_code_patterns' field must be a list of strings. Got foo.
        >>> fn = reorder_measurements({"ordered_code_patterns": [32]}, code_metadata_df)
        Traceback (most recent call last):
            ...
        ValueError: Each element of 'ordered_code_patterns' must be a string. Got 32.
    """

    ordered_code_patterns = stage_cfg.get("ordered_code_patterns", None)
    if not ordered_code_patterns:
        return lambda df: df

    if not isinstance(ordered_code_patterns, list | ListConfig):
        raise ValueError(
            f"The 'ordered_code_patterns' field must be a list of strings. Got {ordered_code_patterns}."
        )

    num_code_patterns = len(ordered_code_patterns)
    code_pattern_idx_dtype = get_smallest_valid_uint_type(num_code_patterns + 1)

    join_cols = ["code"]
    if code_modifiers:
        logger.warning("Code reordering currently only matches against the 'code' column, not code modifiers")
        join_cols.extend(code_modifiers)

    cols_to_select = ["code"]
    if code_modifiers:
        cols_to_select.extend(code_modifiers)

    code_order_idx_exprs = pl  # this let's us chain the when/then/otherwise calls equivalently over the loop.
    for i, code_matcher in enumerate(ordered_code_patterns):
        if not isinstance(code_matcher, str):
            raise ValueError(f"Each element of 'ordered_code_patterns' must be a string. Got {code_matcher}.")
        logger.debug(f"Creating code order index expression @ {i} for code pattern {code_matcher}")
        matcher_expr = pl.col("code").str.contains(code_matcher)
        idx_expr = pl.lit(i, dtype=code_pattern_idx_dtype)
        code_order_idx_exprs = code_order_idx_exprs.when(matcher_expr).then(idx_expr)

    code_order_idx_exprs = code_order_idx_exprs.otherwise(num_code_patterns).alias("code_order_idx")

    code_indices = code_metadata.lazy().select(*join_cols, code_order_idx_exprs)

    def reorder_fn(df: pl.LazyFrame) -> pl.LazyFrame:
        """Reorders the measurements within each event (unique timepoint) by the specified code order."""

        return (
            df.join(code_indices, on=join_cols, how="left", coalesce=True)
            .sort("subject_id", "time", "code_order_idx", maintain_order=True)
            .drop("code_order_idx")
        )

    return reorder_fn



"""Code for manipulating compute functions with respect to configuration or stage configuration objects."""

from __future__ import annotations

import inspect
from collections.abc import Callable
from enum import Enum, auto
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, NotRequired, TypedDict

import polars as pl

from ..dataframe import DF_T

if TYPE_CHECKING:
    from omegaconf import DictConfig

COMPUTE_FN_T = Callable[[DF_T], DF_T]
COMPUTE_FN_UNBOUND_T = Callable[..., DF_T]
COMPUTE_FNTR_T = Callable[..., COMPUTE_FN_T]
ANY_COMPUTE_FN_T = COMPUTE_FN_T | COMPUTE_FN_UNBOUND_T | COMPUTE_FNTR_T


class ComputeFnArgs(TypedDict, total=False):
    df: NotRequired[DF_T]
    dfs: NotRequired[DF_T]
    cfg: NotRequired[DictConfig]
    stage_cfg: NotRequired[DictConfig]
    code_modifiers: NotRequired[list[str]]
    code_metadata: NotRequired[pl.DataFrame]


class ComputeFnType(Enum):
    """Stores the three types of compute functions that can be used in this utility.

    Attributes:
        DIRECT: A direct compute function that maps a DataFrame to a DataFrame.
        UNBOUND: An unbound compute function that needs to be called with additional parameters in addition to
            the input DataFrame, but still returns a DataFrame.
        FUNCTOR: A functor that needs to be called with non "df" args to return a direct compute function.
    """

    DIRECT = auto()
    UNBOUND = auto()
    FUNCTOR = auto()

    @classmethod
    def from_fn(cls, compute_fn: ANY_COMPUTE_FN_T) -> ComputeFnType | None:
        """Returns the type of a compute function or None if invalid.

        Behavior:
          - Returns `ComputeFnType.DIRECT` if the compute function is a direct compute function -- i.e., if it
            takes only a single parameter named `df` or `dfs` and if the return annotation (if one exists) is
            not a `Callable` type annotation.
          - Returns `ComputeFnType.UNBOUND` if the compute function is an unbound compute function -- i.e., if
            it takes a `df` or `dfs` parameter and at least one other parameter among the set of allowed
            additional or functor parameters, and if the return annotation (if one exists) is not a `Callable`
            type annotation.
          - Returns `ComputeFnType.FUNCTOR` if the compute function is a functor that returns a direct compute
            function when called with the appropriate parameters -- i.e., if it takes no `df` parameter, if
            all parameters are among the allowed functor parameters, and if the return annotation (if one
            exists) is a `Callable` type annotation.

        Allowed functor parameters are:
          - `cfg` for the DictConfig configuration object.
          - `stage_cfg` for the DictConfig stage configuration object.
          - `code_modifiers` for a list of code modifier columns.
          - `code_metadata` for a Polars DataFrame of code metadata

        Args:
            compute_fn: The compute function to check.

        Returns:
            The type of the compute function or `None` if invalid.

        Examples:
            >>> def compute_fn(df: pl.DataFrame) -> pl.DataFrame:
            ...     return df
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.DIRECT: 1>
            >>> def compute_fn(*dfs: pl.DataFrame) -> pl.DataFrame:
            ...     return pl.concat(dfs)
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.DIRECT: 1>
            >>> def compute_fn(df: pl.DataFrame) -> dict[Any, list[Any]]:
            ...     return df.to_dict(as_series=False)
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.DIRECT: 1>
            >>> def compute_fn(df: pl.DataFrame):
            ...     return None
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.DIRECT: 1>
            >>> def compute_fn(foo: pl.DataFrame):
            ...     return None
            >>> print(ComputeFnType.from_fn(compute_fn))
            None
            >>> def compute_fn(df: pl.DataFrame, cfg: DictConfig) -> pl.DataFrame:
            ...     return df
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.UNBOUND: 2>
            >>> def compute_fn(df: pl.DataFrame, foo: DictConfig) -> pl.DataFrame:
            ...     return df
            >>> print(ComputeFnType.from_fn(compute_fn))
            None
            >>> def compute_fn(df: pl.LazyFrame, cfg: DictConfig):
            ...     return df
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.UNBOUND: 2>
            >>> def compute_fn(cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
            ...     return lambda df: df
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.FUNCTOR: 3>
            >>> def compute_fn(cfg: DictConfig):
            ...     return lambda df: df
            >>> ComputeFnType.from_fn(compute_fn)
            <ComputeFnType.FUNCTOR: 3>
            >>> def compute_fn(df: pl.DataFrame, cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
            ...     return lambda df: df
            >>> print(ComputeFnType.from_fn(compute_fn))
            None
        """
        sig = inspect.signature(compute_fn)

        allowed_params = {*ComputeFnArgs.__required_keys__, *ComputeFnArgs.__optional_keys__}
        all_params_allowed = all(param in allowed_params for param in sig.parameters)
        if not all_params_allowed:
            return None

        has_df_param = ("df" in sig.parameters) or ("dfs" in sig.parameters)
        has_only_df_param = has_df_param and (len(sig.parameters) == 1)

        return_annotation = sig.return_annotation

        if return_annotation is inspect.Signature.empty:
            has_return_annotation = False
            has_callable_return = False
        elif isinstance(return_annotation, str):
            has_return_annotation = True
            has_callable_return = return_annotation.startswith("Callable[")
        elif hasattr(return_annotation, "__name__"):
            has_return_annotation = return_annotation.__name__ != "_empty"
            has_callable_return = return_annotation.__name__.startswith("Callable")
        else:  # pragma: no cover
            raise ValueError(
                f"Cannot parse return annotation type for {compute_fn.__name__}: {return_annotation}"
            )

        if has_only_df_param:
            if has_return_annotation:
                return cls.DIRECT if not has_callable_return else None
            return cls.DIRECT
        elif has_df_param:
            if has_return_annotation:
                return cls.UNBOUND if not has_callable_return else None
            return cls.UNBOUND
        else:
            if has_return_annotation:
                return cls.FUNCTOR if has_callable_return else None
            return cls.FUNCTOR

        print(sig.return_annotation)


def identity_fn(df: Any) -> Any:
    """A "null" compute function that returns the input DataFrame as is.

    Args:
        df: The input DataFrame.

    Returns:
        The input DataFrame.

    Examples:
        >>> df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> (identity_fn(df) == df).select(pl.all_horizontal(pl.all().all())).item()
        True
    """

    return df


def bind_compute_fn(cfg: DictConfig, stage_cfg: DictConfig, compute_fn: ANY_COMPUTE_FN_T) -> COMPUTE_FN_T:
    """Bind the compute function to the appropriate parameters based on the type of the compute function.

    Args:
        cfg: The DictConfig configuration object.
        stage_cfg: The DictConfig stage configuration object. This is separated from the ``cfg`` argument
            because in some cases, such as under the match & revise paradigm, the stage config may be modified
            dynamically under different matcher conditions to yield different compute functions.
        compute_fn: The compute function to bind.

    Returns:
        The compute function bound to the appropriate parameters.

    Raises:
        ValueError: If the compute function is not a valid compute function.

    Examples:
        >>> compute_fn = bind_compute_fn(DictConfig({}), DictConfig({}), None)
        >>> compute_fn("foobar")
        'foobar'
        >>> def compute_fntr(df: pl.DataFrame, cfg: DictConfig) -> pl.DataFrame:
        ...     return df.with_columns(pl.lit(cfg.val).alias("val"))
        >>> compute_fn = bind_compute_fn(DictConfig({"val": "foo"}), None, compute_fntr)
        >>> compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ val │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 1   ┆ foo │
        │ 2   ┆ foo │
        │ 3   ┆ foo │
        └─────┴─────┘
        >>> def direct_compute_fn(df: pl.DataFrame) -> pl.DataFrame:
        ...     return df.with_columns(pl.lit("bar").alias("val"))
        >>> compute_fn = bind_compute_fn(DictConfig({"val": "foo"}), None, direct_compute_fn)
        >>> compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ val │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 1   ┆ bar │
        │ 2   ┆ bar │
        │ 3   ┆ bar │
        └─────┴─────┘
        >>> def compute_fntr(cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
        ...     return lambda df: df.with_columns(pl.lit(cfg.val).alias("val"))
        >>> compute_fn = bind_compute_fn(DictConfig({"val": "foo"}), None, compute_fntr)
        >>> compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ val │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 1   ┆ foo │
        │ 2   ┆ foo │
        │ 3   ┆ foo │
        └─────┴─────┘
        >>> def compute_fntr(stage_cfg, cfg) -> Callable[[pl.DataFrame], pl.DataFrame]:
        ...     return lambda df: df.with_columns(
        ...         pl.lit(stage_cfg.val).alias("stage_val"), pl.lit(cfg.val).alias("cfg_val")
        ...     )
        >>> compute_fn = bind_compute_fn(DictConfig({"val": "quo"}), DictConfig({"val": "bar"}), compute_fntr)
        >>> compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 3)
        ┌─────┬───────────┬─────────┐
        │ a   ┆ stage_val ┆ cfg_val │
        │ --- ┆ ---       ┆ ---     │
        │ i64 ┆ str       ┆ str     │
        ╞═════╪═══════════╪═════════╡
        │ 1   ┆ bar       ┆ quo     │
        │ 2   ┆ bar       ┆ quo     │
        │ 3   ┆ bar       ┆ quo     │
        └─────┴───────────┴─────────┘
        >>> def compute_fntr(df, code_metadata):
        ...     return df.join(code_metadata, on="a")
        >>> code_metadata_df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> from tempfile import TemporaryDirectory
        >>> with TemporaryDirectory() as tmpdir:
        ...     code_metadata_fp = Path(tmpdir) / "codes.parquet"
        ...     code_metadata_df.write_parquet(code_metadata_fp)
        ...     stage_cfg = DictConfig({"metadata_input_dir": tmpdir})
        ...     compute_fn = bind_compute_fn(DictConfig({}), stage_cfg, compute_fntr)
        ...     compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> def compute_fntr(df: pl.DataFrame, cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
        ...     return lambda df: df
        >>> bind_compute_fn(DictConfig({}), DictConfig({}), compute_fntr)
        Traceback (most recent call last):
            ...
        ValueError: Invalid compute function
    """

    if compute_fn is None:
        return identity_fn

    def fntr_params(compute_fn: ANY_COMPUTE_FN_T) -> ComputeFnArgs:
        compute_fn_params = inspect.signature(compute_fn).parameters
        kwargs = ComputeFnArgs()

        if "cfg" in compute_fn_params:
            kwargs["cfg"] = cfg
        if "stage_cfg" in compute_fn_params:
            kwargs["stage_cfg"] = stage_cfg
        if "code_modifiers" in compute_fn_params:
            code_modifiers = cfg.get("code_modifiers", None)
            kwargs["code_modifiers"] = code_modifiers
        if "code_metadata" in compute_fn_params:
            kwargs["code_metadata"] = pl.read_parquet(
                Path(stage_cfg.metadata_input_dir) / "codes.parquet", use_pyarrow=True
            )
        return kwargs

    match ComputeFnType.from_fn(compute_fn):
        case ComputeFnType.DIRECT:
            return compute_fn
        case ComputeFnType.UNBOUND:
            return partial(compute_fn, **fntr_params(compute_fn))
        case ComputeFnType.FUNCTOR:
            return compute_fn(**fntr_params(compute_fn))
        case _:
            raise ValueError("Invalid compute function")



import subprocess

from MEDS_transforms.stages import StageExample


def test_stage_help(stage: str):
    """Test the help command for a stage."""

    script = f"MEDS_transform-stage __null__ {stage} --help"

    subprocess.run(script, shell=True, check=True)


def test_stage_scenario(stage_example: StageExample):
    stage_example.test()



"""Basic utilities for parallelizable map operations on sharded MEDS datasets with caching and locking."""

import logging
from pathlib import Path

from ..compute_modes import COMPUTE_FN_T
from ..dataframe import READ_FN_T, WRITE_FN_T, read_df, write_df
from .rwlock import rwlock_wrap
from .shard_iteration import InOutFilePair

logger = logging.getLogger(__name__)


def map_over(
    shards: list[InOutFilePair],
    map_fn: COMPUTE_FN_T,
    read_fn: READ_FN_T = read_df,
    write_fn: WRITE_FN_T = write_df,
    do_overwrite: bool = False,
) -> list[Path]:
    """Performs a map operation over a list of input file paths, writing the outputs to output files.

    Internally, for each operation, this uses `rwlock_wrap` to ensure that multiple competing workers do not
    attempt to compute over the same input/output pair at once. This means that this function can be run
    safely in parallel without subdividing the set of shards manually in advance.

    Args:
        shards: List of input-output file pairs to process.
        map_fn: Function to perform the mapping operation on the data. It should take one dataframe
            argument and return one dataframe.
        read_fn: Function to read data from the input file paths. If `None`, the default read function
            reads parquet files.
        write_fn: Function to write the mapped data to the output file path.
        do_overwrite: Should this overwrite an existing out file?

    Returns:
        The list of output files written by this operation.

    Examples:
        >>> def map_fn(df: pl.DataFrame) -> pl.DataFrame:
        ...     return df.with_columns(pl.col("A") * 2)
        >>> dfs = [pl.DataFrame({"A": [i, i + 1, i + 2], "B": [1, 2, 3]}) for i in range(3)]
        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     in_fps = [Path(tmp) / f"input_{i}.parquet" for i in range(3)]
        ...     out_fps = [Path(tmp) / f"output_{i}.parquet" for i in range(3)]
        ...     for fp, df in zip(in_fps, dfs):
        ...         df.write_parquet(fp)
        ...     shards = list(zip(in_fps, out_fps))
        ...     out_fps = map_over(shards, map_fn)
        ...     for fp in out_fps:
        ...         print(fp.relative_to(tmp))
        ...         print(pl.read_parquet(fp))
        output_0.parquet
        shape: (3, 2)
        ┌─────┬─────┐
        │ A   ┆ B   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 0   ┆ 1   │
        │ 2   ┆ 2   │
        │ 4   ┆ 3   │
        └─────┴─────┘
        output_1.parquet
        shape: (3, 2)
        ┌─────┬─────┐
        │ A   ┆ B   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 2   ┆ 1   │
        │ 4   ┆ 2   │
        │ 6   ┆ 3   │
        └─────┴─────┘
        output_2.parquet
        shape: (3, 2)
        ┌─────┬─────┐
        │ A   ┆ B   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 4   ┆ 1   │
        │ 6   ┆ 2   │
        │ 8   ┆ 3   │
        └─────┴─────┘
    """

    all_out_fps = []
    for in_fp, out_fp in shards:
        logger.info(f"Processing {in_fp.resolve()!s} into {out_fp.resolve()!s}")
        rwlock_wrap(
            in_fp,
            out_fp,
            read_fn,
            write_fn,
            compute_fn=map_fn,
            do_overwrite=do_overwrite,
        )
        all_out_fps.append(out_fp)
    return all_out_fps
